"""
Tests for work-arounds to known arXiv API bugs.
"""
import arxiv
import unittest


class TestClient(unittest.TestCase):
    def test_missing_title(self):
        """
        Papers with the title "0" do not have a title element in the Atom feed.

        It's unclear whether other falsey titles (e.g. "False", "null", or empty
        titles) are allowed by arXiv and are impacted by this bug. This may also
        surface for other expected fields (e.g. author names).

        + GitHub issue: https://github.com/lukasschwab/arxiv.py/issues/71
        + Bug report: https://groups.google.com/u/1/g/arxiv-api/c/ORENISrc5gc
        """
        paper_without_title = "2104.12255v1"
        try:
            results = list(arxiv.Search(id_list=[paper_without_title]).results())
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0].get_short_id(), paper_without_title)
        except AttributeError:
            self.fail("got AttributeError fetching paper without title")



from arxiv import arxiv
import os
import shutil
import tempfile
import unittest


class TestDownload(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        self.fetched_result = next(arxiv.Search(id_list=["1605.08386"]).results())

    @classmethod
    def setUp(self):
        self.temp_dir = tempfile.mkdtemp()

    @classmethod
    def tearDown(self):
        shutil.rmtree(self.temp_dir)

    def test_download_from_query(self):
        self.fetched_result.download_pdf(dirpath=self.temp_dir)
        self.assertTrue(os.path.exists(os.path.join(
            self.temp_dir,
            '1605.08386v1.Heat_bath_random_walks_with_Markov_bases.pdf')
        ))

    def test_download_tarfile_from_query(self):
        self.fetched_result.download_source(dirpath=self.temp_dir)
        self.assertTrue(os.path.exists(os.path.join(
            self.temp_dir,
            '1605.08386v1.Heat_bath_random_walks_with_Markov_bases.tar.gz'
        )))

    def test_download_with_custom_slugify_from_query(self):
        fn = 'custom-filename.extension'
        self.fetched_result.download_pdf(dirpath=self.temp_dir, filename=fn)
        self.assertTrue(os.path.exists(os.path.join(self.temp_dir, fn)))



import unittest
from unittest.mock import MagicMock, call, patch
import arxiv
from datetime import datetime, timedelta
from pytest import approx
import feedparser


class TestClient(unittest.TestCase):
    def test_invalid_format_id(self):
        with self.assertRaises(arxiv.HTTPError):
            list(arxiv.Client(num_retries=0).results(arxiv.Search(id_list=["abc"])))

    def test_invalid_id(self):
        results = list(arxiv.Search(id_list=["0000.0000"]).results())
        self.assertEqual(len(results), 0)

    def test_nonexistent_id_in_list(self):
        # Assert _from_feed_entry throws MissingFieldError.
        feed = feedparser.parse("http://export.arxiv.org/api/query?id_list=0808.05394")
        with self.assertRaises(arxiv.Result.MissingFieldError):
            arxiv.Result._from_feed_entry(feed.entries[0])
        # Assert thrown error is handled and hidden by generator.
        results = list(arxiv.Search(id_list=["0808.05394"]).results())
        self.assertEqual(len(results), 0)
        # Generator should still yield valid entries.
        results = list(arxiv.Search(id_list=["0808.05394", "1707.08567"]).results())
        self.assertEqual(len(results), 1)

    def test_max_results(self):
        client = arxiv.Client(page_size=10, delay_seconds=0)
        search = arxiv.Search(query="testing", max_results=2)
        results = [r for r in client.results(search)]
        self.assertEqual(len(results), 2)

    def test_query_page_count(self):
        client = arxiv.Client(page_size=10, delay_seconds=0)
        client._parse_feed = MagicMock(wraps=client._parse_feed)
        generator = client.results(arxiv.Search(query="testing", max_results=55))
        results = [r for r in generator]
        self.assertEqual(len(results), 55)
        self.assertEqual(client._parse_feed.call_count, 6)

    def test_offset(self):
        max_results = 10
        search = arxiv.Search(query="testing", max_results=max_results)
        client = arxiv.Client(page_size=10, delay_seconds=0)

        default = list(client.results(search))
        no_offset = list(client.results(search))
        self.assertListEqual(default, no_offset)

        offset = max_results // 2
        half_offset = list(client.results(search, offset=offset))
        self.assertListEqual(default[offset:], half_offset)

        offset_above_max_results = list(client.results(search, offset=max_results))
        self.assertListEqual(offset_above_max_results, [])

    def test_search_results_offset(self):
        search = arxiv.Search(query="testing", max_results=10)
        client = arxiv.Client(page_size=3, delay_seconds=0)
        for offset in [0, 5, 9, 10, 11]:
            client_results = list(client.results(search, offset=offset))
            search_results = list(search.results(offset=offset))
            self.assertListEqual(client_results, search_results)

    def test_no_duplicates(self):
        search = arxiv.Search("testing", max_results=100)
        ids = set()
        for r in search.results():
            self.assertFalse(r.entry_id in ids)
            ids.add(r.entry_id)

    @patch('time.sleep', return_value=None)
    def test_retry(self, patched_time_sleep):
        broken_client = TestClient.get_broken_client()

        def broken_get():
            search = arxiv.Search(query="quantum")
            return next(broken_client.results(search))
        self.assertRaises(arxiv.HTTPError, broken_get)

        for num_retries in [2, 5]:
            broken_client.num_retries = num_retries
            try:
                broken_get()
                self.fail("broken_get didn't throw HTTPError")
            except arxiv.HTTPError as e:
                self.assertEqual(e.status, 500)
                self.assertEqual(e.retry, broken_client.num_retries)

    @patch('time.sleep', return_value=None)
    def test_sleep_standard(self, patched_time_sleep):
        client = arxiv.Client(page_size=1)
        url = client._format_url(arxiv.Search(query="quantum"), 0, 1)
        # A client should sleep until delay_seconds have passed.
        client._parse_feed(url)
        patched_time_sleep.assert_not_called()
        # Overwrite _last_request_dt to minimize flakiness: different
        # environments will have different page fetch times.
        client._last_request_dt = datetime.now()
        client._parse_feed(url)
        patched_time_sleep.assert_called_once_with(
            approx(client.delay_seconds, rel=1e-3)
        )

    @patch('time.sleep', return_value=None)
    def test_sleep_multiple_requests(self, patched_time_sleep):
        client = arxiv.Client(page_size=1)
        url1 = client._format_url(arxiv.Search(query="quantum"), 0, 1)
        url2 = client._format_url(arxiv.Search(query="testing"), 0, 1)
        # Rate limiting is URL-independent; expect same behavior as in
        # `test_sleep_standard`.
        client._parse_feed(url1)
        patched_time_sleep.assert_not_called()
        client._last_request_dt = datetime.now()
        client._parse_feed(url2)
        patched_time_sleep.assert_called_once_with(
            approx(client.delay_seconds, rel=1e-3)
        )

    @patch('time.sleep', return_value=None)
    def test_sleep_elapsed(self, patched_time_sleep):
        client = arxiv.Client(page_size=1)
        url = client._format_url(arxiv.Search(query="quantum"), 0, 1)
        # If _last_request_dt is less than delay_seconds ago, sleep.
        client._last_request_dt = (
            datetime.now() - timedelta(seconds=client.delay_seconds-1)
        )
        client._parse_feed(url)
        patched_time_sleep.assert_called_once()
        patched_time_sleep.reset_mock()
        # If _last_request_dt is at least delay_seconds ago, don't sleep.
        client._last_request_dt = (
            datetime.now() - timedelta(seconds=client.delay_seconds)
        )
        client._parse_feed(url)
        patched_time_sleep.assert_not_called()

    @patch('time.sleep', return_value=None)
    def test_sleep_zero_delay(self, patched_time_sleep):
        client = arxiv.Client(page_size=1, delay_seconds=0)
        url = client._format_url(arxiv.Search(query="quantum"), 0, 1)
        client._parse_feed(url)
        client._parse_feed(url)
        patched_time_sleep.assert_not_called()

    @patch('time.sleep', return_value=None)
    def test_sleep_between_errors(self, patched_time_sleep):
        client = TestClient.get_broken_client()
        url = client._format_url(arxiv.Search(query="quantum"), 0, 1)
        try:
            client._parse_feed(url)
        except arxiv.HTTPError:
            pass
        # Should sleep between retries.
        patched_time_sleep.assert_called()
        self.assertEqual(patched_time_sleep.call_count, client.num_retries)
        patched_time_sleep.assert_has_calls([
            call(approx(client.delay_seconds, rel=1e-3)),
        ] * client.num_retries)

    def get_broken_client():
        """
        get_broken_client returns an arxiv.Client that always encounters a 500
        status.
        """
        # TODO: reimplement broken_client with a mock.
        broken_client = arxiv.Client(page_size=1)
        broken_client.query_url_format = "https://httpstat.us/500?{}"
        return broken_client

    def get_once_client():
        """
        get_once_client returns an arxiv.Client that only tries once.
        """
        return arxiv.Client(num_retries=0)



import arxiv
import unittest
import re
import time
from datetime import datetime, timezone


class TestResult(unittest.TestCase):

    def assert_nonempty(self, s):
        self.assertIsNotNone(s)
        self.assertNotEqual(s, '')

    def assert_valid_author(self, a: arxiv.Result.Author):
        self.assert_nonempty(a.name)

    def assert_valid_link(self, link: arxiv.Result.Link):
        self.assert_nonempty(link.href)

    def assert_valid_result(self, result: arxiv.Result):
        self.assert_nonempty(result.entry_id)
        self.assertIsNotNone(result.updated)
        self.assertIsNotNone(result.published)
        self.assert_nonempty(result.title)
        self.assertNotEqual(len(result.authors), 0)
        for author in result.authors:
            self.assert_valid_author(author)
        self.assert_nonempty(result.summary)
        self.assert_nonempty(result.primary_category)
        self.assertNotEqual(len(result.categories), 0)
        for category in result.categories:
            self.assert_nonempty(category)
        for link in result.links:
            self.assert_valid_link(link)
        self.assert_nonempty(result.pdf_url)

    def test_result_shape(self):
        max_results = 100
        search = arxiv.Search("testing", max_results=max_results)
        results = [r for r in search.results()]
        self.assertEqual(len(results), max_results)
        for result in results:
            self.assert_valid_result(result)

    def test_from_feed_entry(self):
        feed = arxiv.Client()._parse_feed(
            "http://export.arxiv.org/api/query?search_query=testing"
        )
        feed_entry = feed.entries[0]
        result = arxiv.Result._from_feed_entry(feed_entry)
        self.assert_valid_result(result)

    def test_get_short_id(self):
        result_id = "1707.08567"
        result = next(arxiv.Search(id_list=[result_id]).results())
        got = result.get_short_id()
        self.assertTrue(got.startswith(result_id))
        # Should be of form `1707.08567v1`.
        self.assertTrue(re.match(r'^{}v\d+$'.format(result_id), got))

    def test_to_datetime(self):
        """Test time.struct_time to datetime conversion."""
        # paper_published and paper_published_parsed correspond to
        # r._raw.published and r._raw.published_parsed for 1605.08386v1. It's
        # critical to the test that they remain equivalent.
        paper_published = '2016-05-26T17:59:46Z'
        paper_published_parsed = time.struct_time((2016, 5, 26, 17, 59, 46, 3, 147, 0))
        expected = datetime(2016, 5, 26, hour=17, minute=59, second=46, tzinfo=timezone.utc)
        actual = arxiv.Result._to_datetime(paper_published_parsed)
        self.assertEqual(actual, expected)
        self.assertEqual(actual.strftime("%Y-%m-%dT%H:%M:%SZ"), paper_published)

    def test_eq(self):
        # Results
        id = "some-result"
        result = arxiv.Result(entry_id=id)
        self.assertTrue(result == result)
        self.assertTrue(result == arxiv.Result(entry_id=id))
        self.assertTrue(arxiv.Result(entry_id=id) == result)
        self.assertFalse(result == arxiv.Result(entry_id="other"))
        self.assertFalse(result == id)
        # Authors
        name = "some-name"
        author = arxiv.Result.Author(name)
        self.assertTrue(author == author)
        self.assertTrue(author == arxiv.Result.Author(name))
        self.assertTrue(arxiv.Result.Author(name) == author)
        self.assertFalse(author == arxiv.Result.Author("other"))
        self.assertFalse(author == id)
        # Links
        href = "some-href"
        link = arxiv.Result.Link(href)
        self.assertTrue(link == link)
        self.assertTrue(link == arxiv.Result.Link(href))
        self.assertTrue(arxiv.Result.Link(href) == link)
        self.assertFalse(link == arxiv.Result.Link("other"))
        self.assertFalse(link == id)

    def test_legacy_ids(self):
        full_legacy_id = "quant-ph/0201082v1"
        result = next(arxiv.Search(id_list=[full_legacy_id]).results())
        self.assertEqual(result.get_short_id(), full_legacy_id)
